#! /u/feeley/icfp2002/g4/bin/gsi -:d

; file: "learn.scm"

(declare
 (standard-bindings)
 (extended-bindings)
 (fixnum)
 (not safe)
 (block)
)

;------------------------------------------------------------------------------

(##include "rand.scm")

;------------------------------------------------------------------------------

(define (main . args)

  (define (usage)
    (display "usage: learn.scm <server-program> <robot-program> <population-size>\n")
    (exit 70))

  (if (< (length args) 3)
      (usage)
      (let* ((server-program (car args))
             (robot-program (cadr args))
             (population-size (string->number (caddr args))))
        (if (or (not population-size)
                (not (integer? population-size))
                (not (exact? population-size)))
            (usage)
            (begin
              (learn server-program
                     robot-program
                     population-size)
              (exit 0))))))

(define server-program "./server")
(define robot-program "./robot")
(define population-size 50)

(define (learn s-program r-program pop-size)
  (set! server-program s-program)
  (set! robot-program r-program)
  (set! population-size pop-size)
  (let ((learnable-parameter-defs
         (get-learnable-parameter-defs)))

    (display "PARAMETERS TO LEARN:\n")
    (pp learnable-parameter-defs)

    (process-defs learnable-parameter-defs)

    (evolve
     (create-initial-population population-size))))

(define-type param
  (name)
  (lo)
  (hi)
  (start)
  (size)
)

(define (process-defs learnable-parameter-defs)
  (let loop ((lst1 learnable-parameter-defs)
             (lst2 '())
             (start 0))
    (if (pair? lst1)
        (let* ((x (car lst1))
               (name (car x))
               (lo (cadr x))
               (hi (caddr x))
               (size (range-size lo hi)))
          (loop (cdr lst1)
                (cons (make-param name
                                  lo
                                  hi
                                  start
                                  size)
                      lst2)
                (+ start size)))
        (begin
          (set! params (reverse lst2))
          (set! gene-size start)
          ))))

(define (range-size lo hi)
  (integer-length (- hi lo)))

(define params #f)
(define gene-size #f)

(define (random-gene)
  (let loop ((n gene-size)
             (gene 0))
    (if (> n 0)
        (loop (- n 1)
              (+ (* 2 gene) (rand 2)))
        gene)))

(define (individual->learnable-parameter-values individual)
  (let ((gene (individual-gene individual)))
    (map (lambda (p)
           (cons (param-name p)
                 (+ (param-lo p)
                    (extract-bit-field (param-size p)
                                       (param-start p)
                                       gene))))
         params)))

(define-type individual
  (gene)
)

(define (create-initial-population population-size) ; population-size must be even
  (let loop ((i (- population-size 1))
             (lst '()))
    (if (< i 0)
        lst
        (loop (- i 1)
              (cons (make-individual
                     (random-gene))
                    lst)))))

(define (crossover! individual1 individual2)
  (let* ((crossover-points
          (let ((nb-crossovers (+ 1 (quotient gene-size 5))))
            (sort-list
             (map (lambda (i) (rand gene-size))
                  (iota nb-crossovers))
             <)))
         (start-crossing-at-0?
          (= (rand 2) 0)))
    (let loop ((g1 (individual-gene individual1))
               (g2 (individual-gene individual2))
               (pos (if start-crossing-at-0? 0 (car crossover-points)))
               (lst (if start-crossing-at-0? crossover-points (cdr crossover-points))))
      (let* ((end (if (pair? lst) (car lst) gene-size))
             (size (- end pos)))
        (let* ((new-g1
                (insert-bit-field size
                                  pos
                                  (extract-bit-field size
                                                     pos
                                                     g2)
                                  g1))
               (new-g2
                (insert-bit-field size
                                  pos
                                  (extract-bit-field size
                                                     pos
                                                     g1)
                                  g2)))
          (if (and (pair? lst) (pair? (cdr lst)))
              (loop new-g1
                    new-g2
                    (cadr lst)
                    (cddr lst))
              (cons
               (make-individual new-g1)
               (make-individual new-g2))))))))

(define skew 0.8)

(define (skewed-rand n)
  (let ((r (random)))
    (let ((a (* 2.0 skew))
          (b (* 2.0 (- 1.0 skew)))
          (c (* -2.0 r)))
      (let ((x (/ (+ (- b) (sqrt (- (* b b) (* 4.0 a c)))) (* 2.0 a))))
        (inexact->exact (floor (* x (exact->inexact n))))))))

(define (evolve population)
  (let loop1 ((generation 0)
              (population population))

    (pp '--------------------------------------------------)

    (let ((e (evaluate population)))
      (let* ((ordered-population
              (list->vector
               (sort-list e (lambda (x y) (< (car x) (car y))))))
             (n
              (vector-length ordered-population))
             (best
              (vector-ref ordered-population (- n 1))))

        (pp (list 'at-generation: generation
                  'best: (individual->learnable-parameter-values (cdr best))
                  'score: (car best)))

        (pp ordered-population);;;;;;;;;;;;

        (if (< generation 100)

            (loop1 (+ generation 1)
                   (breed
                    (list->vector
                     (map cdr
                          (vector->list ordered-population))))))))))

(define (breed ordered-population)
  (let ((n (vector-length ordered-population)))
    (let loop2 ((i 0)
                (lst '()))
      (if (< i n)
          (let* ((i1
                  (if (= i 0)
                      (vector-ref ordered-population (- n 1))
                      (vector-ref ordered-population (skewed-rand n))))
                 (i2
                  (if (= i 0)
                      (vector-ref ordered-population (- n 2))
                      (vector-ref ordered-population (skewed-rand n)))))
            (if (eq? i1 i2)
                (loop2 i
                       lst)
                (let ((x (crossover! i1 i2)))
                  (loop2 (+ i 2)
                         (cons (car x) (cons (cdr x) lst))))))
          (begin
            (pp lst)
            lst)))))

(define (evaluate population)
  (let ((player-values
         (map individual->learnable-parameter-values population)))
    (pp (list 'trial: (car set-of-trials)))
    (let loop ((lst
                (cdr set-of-trials))
               (scores
                (apply
                 run-trial
                 (append (car set-of-trials) (list player-values)))))
      (if (pair? lst)
          (begin
            (pp (list 'trial: (car lst)))
            (let ((other-scores
                   (apply
                    run-trial
                    (append (car lst) (list player-values)))))
              (loop (cdr lst)
                    (map + scores other-scores))))
          (map cons scores population)))))

(define set-of-trials
  '(                 ; money capacity packages
;;;    ("../../test/1"    2000  100      100)
;    ("../../test/2"    2000  100      100)
;;;    ("../../test/3"    2000  100      100)
;    ("../../test/4"    2000  100      100)
;;;    ("../../test/5"    2000  100      100)
;;;    ("../../test/6"    2000  100      100)
    ("../../test/7"    2000  100      100)
;    ("../../test/8"    2000  100      100)
;;;    ("../../test/9"    2000  100      100)
   )
)

;------------------------------------------------------------------------------

(define foldr
  (lambda (f base lst)
    (if (null? lst)
        base
        (f (car lst) (foldr f base (cdr lst))))))

(define foldl
  (lambda (f base lst)
    (if (null? lst)
        base
        (foldl f (f base (car lst)) (cdr lst)))))

(define sum
  (lambda (lst)
    (foldl + 0 lst)))

(define product
  (lambda (lst)
    (foldl * 1 lst)))

;------------------------------------------------------------------------------

(define (sort-list l <?)

  (define (mergesort l)

    (define (merge l1 l2)
      (cond ((null? l1) l2)
            ((null? l2) l1)
            (else
             (let ((e1 (car l1)) (e2 (car l2)))
               (if (<? e1 e2)
                 (cons e1 (merge (cdr l1) l2))
                 (cons e2 (merge l1 (cdr l2))))))))

    (define (split l)
      (if (or (null? l) (null? (cdr l)))
        l
        (cons (car l) (split (cddr l)))))

    (if (or (null? l) (null? (cdr l)))
      l
      (let* ((l1 (mergesort (split l)))
             (l2 (mergesort (split (cdr l)))))
        (merge l1 l2))))

  (mergesort l))

(define (iota n)
  (let loop ((i (- n 1)) (lst '()))
    (if (< i 0)
        lst
        (loop (- i 1) (cons i lst)))))

;------------------------------------------------------------------------------

(define (run-trial world-file money capacity packages player-values)
  (let ((port-num 5002)
        (nb-players (length player-values))
        (trace? #f))
    (let ((temp (unique-filename)))
      (let ((whole-command
             (string-append
              (background
               (map (lambda (learnable-parameter-values robot-id)
                      (build-robot-command robot-program
                                           "localhost"
                                           port-num
                                           robot-id
                                           learnable-parameter-values))
                    player-values
                    (iota nb-players)))
              " ; ("
              (build-server-command server-program
                                    port-num
                                    nb-players
                                    world-file
                                    money
                                    capacity
                                    packages
                                    trace?)
              " > "
              temp
              ")")))

;        (display "EXECUTING:\n")
;        (display whole-command)
;        (newline)

        (shell-command whole-command)
        (let ((result (with-input-from-file temp read)))
          (delete-file temp)
          result)))))

(define (get-learnable-parameter-defs)
  (let ((temp (unique-filename)))
    (let ((whole-command
           (string-append robot-program
                          " dummy 9999 learnable-parameter-defs > "
                          temp)))
      (shell-command whole-command)
      (let ((result (with-input-from-file temp read)))
        (delete-file temp)
        result))))

;------------------------------------------------------------------------------

(define (build-server-command server-program
                              port-num
                              nb-players
                              world-file
                              money
                              capacity
                              packages
                              trace?)
  (string-append
   server-program
   " "
   (number->string port-num)
   " "
   (number->string nb-players)
   " "
   world-file
   " "
   (number->string money)
   " "
   (number->string capacity)
   " "
   (number->string packages)
   " "
   " \""
   (with-output-to-string (lambda () (write trace?)))
   "\""))

(define (build-robot-command robot-program
                             server-hostname
                             port-num
                             robot-id
                             learnable-parameter-values)
  (string-append
   "(sleep 1 ; "
   robot-program
   " "
   server-hostname
   " "
   (number->string port-num)
   " "
   (number->string robot-id)
   " \""
   (with-output-to-string (lambda () (write learnable-parameter-values)))
   "\" < "
   robot-program
   " > temp.robot"
   (number->string robot-id)
   " 2>&1)"))

(define (background commands)

  (define (build-single-command commands)
    (cond ((null? commands)
           "")
          (else
           (string-append
            "("
            (car commands)
            " &)"
            (if (null? (cdr commands))
                ""
                (string-append " ; "
                               (build-single-command (cdr commands))))))))

  (build-single-command commands))

(define (unique-filename)
  (string-append
   "temp"
   (number->string (time->seconds (current-time)))))

;------------------------------------------------------------------------------

(apply main (cdr (command-line)))

(force-output (current-output-port)) ; **** IMPORTANT ****
