#! /u/feeley/icfp2002/g4/bin/gsi -:d

; file: "runme.scm"

(declare
 (standard-bindings)
 (extended-bindings)
 (fixnum)
 (not safe)
 (block)
)

;******** make sure the value of the learnable parameters that
;******** were discovered by the genetic algorithm are put here
;******** before the final compilation!!!!!

(define learnable-parameter-values
  '(
    (fancy-dist-map . 1)
    (enemy-cost . 1)
    (drowning-cost . 3)
    (unvisited-cost . 2)
    (empty-cost . 0)
    (asociability-min-dist . 3)
   ))

;------------------------------------------------------------------------------

; For learning parameters:

(define learnable-parameter-defs '())

(define (make-learnable-parameter name lo hi)
  ; parameter will be in the interval lo..hi
  (set! learnable-parameter-defs
        (append learnable-parameter-defs
                (list (list name lo hi))))
  (let ((value
         (let ((x (assoc name learnable-parameter-values)))
           (if x
               (cdr x)
               (quotient (+ lo hi) 2)))))
    (lambda () value)))

(define-macro (define-learnable-parameter name lo hi)
  `(define ,name (make-learnable-parameter ',name ,lo ,hi)))

;------------------------------------------------------------------------------

(define-type robot
  (id)
  (pos)
  (packages))

(define-type package
  (id)
  (pos)
  (weight))

;------------------------------------------------------------------------------

;;;(##set-min-heap! 32000000)

(##include "visu.scm")

(define (main . args)

  (define (usage)
    (display "usage: runme.scm <server-hostname> <port-num> [<other-options>...]\n")
    (display "  <other-options> = learnable-parameter-defs\n")
    (display "                  | visualize\n")
    (display "                  | trace\n")
    (display "                  | ((asociability-min-dist . 5) ...)\n")
    (display "                  | <robot-id>\n")
    (exit 70))

  (if (< (length args) 2)
      (usage)
      (let* ((server-hostname (car args))
             (port-num (string->number (cadr args))))
        (if (or (not port-num)
                (not (integer? port-num))
                (not (exact? port-num)))
            (usage)
            (let ((other-options
                   (map (lambda (str)
                          (read (open-input-string str)))
                        (cddr args))))
              (let ((x (keep list? other-options)))
                (if (pair? x)
                    (set! learnable-parameter-values (car x))))
              (if (memq 'learnable-parameter-defs other-options)
                  (begin
                    (pp learnable-parameter-defs)
                    (force-output)
                    (exit 0))
                  (begin
                    (set! visualize? (memq 'visualize other-options))
                    (set! trace? (memq 'trace other-options))
                    (if trace?
                        (pp (list 'learnable-parameter-values
                                  '=
                                  learnable-parameter-values)))
                    (start-client server-hostname
                                  port-num
                                  (let ((x (keep integer? other-options)))
                                    (if (pair? x)
                                        (car x)
                                        #f))))))))))

(define (start-client server-hostname port-num id)

  (set! package-ids '())
  (set! packages (make-table))

  (set! nb-steps 1)

  (let ((port (open-tcp-client server-hostname port-num)))

    ; 1 - Client sends the string "Player"

    (if id
        (begin
          (display "Player " port)
          (display id port)
          (newline port))
        (display "Player\n" port))
    (force-output port)

    ; 2 - Server sends board, player's configuration, and initial response

    (read-board port)
    (read-conf port)
    (process-response (read-response port) #t)

    (set! me (id->robot rid))

    (robot-packages-set! me '()) ; we are sure we have no packages

    (set! control-map
     (compute-control-map! robot-ids))

    (set! control-map-excluding-me
     (compute-control-map!
      (keep (lambda (id) (not (= id rid))) robot-ids)))

    (set! distance-map-from-me
     (compute-distance-map (robot-pos me)))

    (set! bases (bases-reachable-from (robot-pos me)))
    (compute-base-distance-map)

    (if visualize?
        (visu:init!
         (string-append "View of the world of ROBOT " (number->string rid))
         width+2
         height+2))

    ; Game play

    (let loop ()

;      (if (or (obstacle? (robot-pos me))       ; Detection correcte de la mort
;	      (< money 0))
;	  (exit 0))
;;; OUAIS, tu parles!

      (read-packages port)

      (set! control-map
       (compute-control-map! robot-ids))

      (set! control-map-excluding-me
       (compute-control-map!
        (keep (lambda (id) (not (= id rid))) robot-ids)))

      (set! distance-map-from-me
       (compute-distance-map (robot-pos me)))

      (if (not RO-path)
	  (set! RO-path (cleanup-RO-path (visit-and-deliver))))

      (think-and-send-command port)

      (process-response (read-response port) #f)

      (if visualize?
          (visu:draw-all!))

      (set! nb-steps (+ nb-steps 1))

      (loop))))

(define (dump-state)
  (pp '------------------------------)
  (pp (list 'rid: rid))
  (pp (list 'capacity: capacity))
  (pp (list 'money: money))
  (write board) (newline)
  (pp (list 'bases: bases))
  (pp (list 'robots: robots))
  (pp (list 'packages: packages))
)

(define (read-packages port)
  (let ((line (read-line port)))
    (if trace?
        (begin
          (display "==> ")
          (display line)
          (newline)))
    (if (eof-object? line)            ; Detection correcte de la mort (take II)
	(begin
	  (close-port port)
	  (force-output)
	  (exit 0)))
    (if (not (string? line))
        (error "expected list of available packages")
        (let ((x (read-all (open-input-string line))))
          (if (or (not (list? x))
                  (not (= (modulo (length x) 4) 0)))
              (error "malformed package list was sent by server")
              (let loop ((lst1 x) (lst2 '()))
                (if (pair? lst1)
                    (let* ((id
                            (car lst1))
                           (p
                            (make-package
                             id
                             (col-row->pos (cadr lst1) (caddr lst1))
                             (cadddr lst1))))
                      (if (not (id->package id))
                          (set! package-ids (cons id package-ids)))
                      (table-set! packages id p)
                      (loop (cddddr lst1)
                            (cons p
                                  lst2)))
                    (let ((pos (robot-pos me)))
                      (set-packages-at! pos (map package-id lst2))
                      ))))))))

(define (read-board port)
  (let ((line (read-line port)))
    (if (not (string? line))
        (error "expected board dimensions from server")
        (let* ((p (open-input-string line))
               (width (read p))
               (height (read p)))
          (if (or (not (integer? width))
                  (not (exact? width))
                  (< width 1)
                  (not (integer? height))
                  (not (exact? height))
                  (< height 1))
              (error "dimensions sent by the server are not positive integers")
              (begin
                (set! width+2 (+ width 2))
                (set! height+2 (+ height 2))
                (set! N width+2)
                (set! S (- width+2))
                (set! board (make-vector (* height+2 width+2) (make-wall)))
                (set! bases '())
                (let loop1 ((row 1))
                  (if (<= row height)
                      (let ((line (read-line port)))
                        (if (or (not (string? line))
                                (not (= (string-length line) width)))
                            (error "malformed board was sent by the server")
                            (begin
                              (let loop2 ((col 1))
                                (if (<= col width)
                                    (let ((c (string-ref line (- col 1)))
                                          (pos (col-row->pos col row)))
                                      (if (char=? c #\@)
                                          (set! bases
                                                (cons pos
                                                      bases)))
                                      (vector-set! board pos
                                       (translate-cell c))
                                      (loop2 (+ col 1)))
                                    (loop1 (+ row 1)))))))))))))))

(define (read-conf port)
  (let ((line (read-line port)))
    (if (not (string? line))
        (error "expected configuration from server")
        (let ((x (read-all (open-input-string line))))
          (if (or (not (list? x))
                  (not (= (length x) 3)))
              (error "malformed configuration was sent by server")
              (begin
                (set! rid (car x))
                (set! capacity (cadr x))
                (set! money (caddr x))))))))

(define (read-response port)
  (let ((line (read-line port)))
    (if trace?
        (begin
          (display "==> ")
          (display line)
          (newline)))
    (if (not (string? line))
        (error "expected response from server")
        line)))

(define (process-response response initial?)
  (let* ((lst
          (read-all (open-input-string response)
                    (lambda (p) (read-line p #\#))))
         (live-robot-ids
          '())
         (live-robots
          (make-table)))

    (for-each
     (lambda (id)
       (let ((r (id->robot id)))
         (remove-robot-at! (robot-pos r))))
     robot-ids)

    (if (or (null? lst)
            (not (string=? "" (car lst))))
        (error "malformed server response")
        (for-each
         (lambda (str)
           (let ((lst2 (read-all (open-input-string str))))
             (if (or (null? lst2)
                     (not (integer? (car lst2)))
                     (not (exact? (car lst2)))
                     (< (car lst2) 0))
                 (error "malformed server response entry")
                 (let ((id (car lst2)))
                   (let loop ((lst3 (cdr lst2))
                              (moves '())
                              (picked '())
                              (dropped '())
                              (x #f)
                              (y #f))
                     (if (pair? lst3)
                         (let* ((action (car lst3))
                                (rest (cdr lst3)))
                           (case action
                             ((N S E W)
                              (loop rest
                                    (cons action moves)
                                    picked
                                    dropped
                                    x
                                    y))
                             ((P)
                              (if (not (pair? rest))
                                  (error "P expects an argument")
                                  (loop (cdr rest)
                                        moves
                                        (cons (car rest) picked)
                                        dropped
                                        x
                                        y)))
                             ((D)
                              (if (not (pair? rest))
                                  (error "D expects an argument")
                                  (loop (cdr rest)
                                        moves
                                        picked
                                        (cons (car rest) dropped)
                                        x
                                        y)))
                             ((X)
                              (if (not (pair? rest))
                                  (error "X expects an argument")
                                  (loop (cdr rest)
                                        moves
                                        picked
                                        dropped
                                        (car rest)
                                        y)))
                             ((Y)
                              (if (not (pair? rest))
                                  (error "Y expects an argument")
                                  (loop (cdr rest)
                                        moves
                                        picked
                                        dropped
                                        x
                                        (car rest))))
                             (else
                              (error "expected one of: N S E W P D X Y"))))
                         (begin
                           (if initial?
                               (begin
                                 (if (or (not (null? moves))
                                         (not (null? picked))
                                         (not (null? dropped))
                                         (not x)
                                         (not y))
                                     (error "initial response should only contain X and Y")
                                     (let ((pos (col-row->pos x y)))
                                       (set! live-robot-ids
                                             (cons id live-robot-ids))
                                       (table-set!
                                        live-robots
                                        id
                                        (make-robot
                                         id
                                         pos
                                         '(uncertain))))))
                               (let ((r
                                      (or (id->robot id)
                                          (make-robot
                                           id
                                           0
                                           '(uncertain)))))
                                 (set! live-robot-ids
                                       (cons id live-robot-ids))
                                 (table-set! live-robots id r)

                                 (if (and x y)
                                     (let ((pos (col-row->pos x y)))
                                       (robot-pos-set! r pos)))

                                 (let ((pos (robot-pos r)))

                                   (if moves
                                       (let loop ((pos pos)
                                                  (lst (reverse moves)))
                                         (if (pair? lst)
                                             (loop
                                              (displace pos (car lst))
                                              (cdr lst))
                                             (robot-pos-set! r pos))))

                                   (if (not (null? picked))
                                       (let ((packages (robot-packages r)))
                                         (robot-packages-set! r
                                          (list-union packages picked))
                                         (set-packages-at! pos
                                          (list-diff (pos->packages pos)
                                                     picked))))

                                   (if (not (null? dropped))
                                       (let ((packages (robot-packages r)))
                                         (robot-packages-set! r
                                          (list-diff packages dropped))
                                         ;;;;;;;; if "pos" is the package's
                                         ;;;;;;;; destination then it
                                         ;;;;;;;; vanishes from the board!!
                                         (set-packages-at! pos
                                          (->uncertain (pos->packages pos)))))

                                   ))))))))))
         (cdr lst)))

    (set! robot-ids live-robot-ids)
    (set! robots live-robots)

    (for-each
     (lambda (id)
       (let ((r (id->robot id)))
         (set-robot-at! (robot-pos r) r)))
     robot-ids)))

(define (list-union lst1 lst2)
  (if (pair? lst1)
      (if (memq (car lst1) lst2)
          (list-union (cdr lst1) lst2)
          (cons (car lst1) (list-union (cdr lst1) lst2)))
      lst2))

(define (list-diff lst1 lst2)
  (if (pair? lst1)
      (if (memq (car lst1) lst2)
          (list-diff (cdr lst1) lst2)
          (cons (car lst1) (list-diff (cdr lst1) lst2)))
      '()))

(define (displace pos move)
  (case move
    ((N)
     (+ pos N))
    ((S)
     (+ pos S))
    ((E)
     (+ pos E))
    ((W)
     (+ pos W))
    (else
     (error "unknown move" move))))

(define (uncertain? x)
  (and (pair? x) (eq? (car x) 'uncertain)))

(define (->uncertain x)
  (if (uncertain? x)
      x
      (cons 'uncertain x)))

(define (->certain x)
  (if (uncertain? x)
      (cdr x)
      x))

(define (pos->packages pos)
  (if (or (< pos 0)
          (not (< pos (vector-length board))))
      '()
      (let ((x (vector-ref board pos)))
        (cond ((##fixnum? x)
               '())
              ((and (pair? x) (robot? (car x)))
               (cdr x))
              (else
               x)))))

(define (pos->robot pos)
  (if (or (< pos 0)
          (not (< pos (vector-length board))))
      #f
      (let ((x (vector-ref board pos)))
        (cond ((##fixnum? x)
               #f)
              ((and (pair? x) (robot? (car x)))
               (car x))
              (else
               #f)))))

(define (set-packages-at! pos packages)
  (let ((x (vector-ref board pos)))
    (if (not (##fixnum? x))
	(if (and (pair? x) (robot? (car x)))
	    (vector-set! board pos (cons (car x) packages))
	    (vector-set! board pos packages)))))

(define (set-robot-at! pos rob)
  (let ((x (vector-ref board pos)))
    (if (not (##fixnum? x))
	(if (and (pair? x) (robot? (car x)))
	    (vector-set! board pos (cons rob (cdr x)))
	    (vector-set! board pos (cons rob x))))))

(define (remove-robot-at! pos)
  (let ((x (vector-ref board pos)))
    (if (not (##fixnum? x))
	(if (and (pair? x) (robot? (car x)))
	    (vector-set! board pos (cdr x))
	    #f))))

'
(define (pos->robot-old pos)
(let ((q
  (let loop ((lst robot-ids))
    (if (pair? lst)
        (let* ((id (car lst))
               (r (id->robot id)))
          (if (= (robot-pos r) pos)
              r
              (loop (cdr lst))))
        #f)))
) (let ((qq (pos->robot2 pos))) (if (not (eq? q qq)) (error "not-same" q qq))
q)))

(define (id->robot id)
  (table-ref robots id))

(define (id->package id)
  (table-ref packages id))

; Board representation:

(define (make-wall) 1)
(define (make-water) 0)
(define (make-empty) '(uncertain)) ; list of packages known to be here
                                   ; when "uncertain" is at the head of the
                                   ; list there may be other packages,
                                   ; otherwise the list is complete

(define (wall? pos) (eq? (vector-ref board pos) 1))
(define (water? pos) (eq? (vector-ref board pos) 0))
(define (obstacle? pos) (##fixnum? (vector-ref board pos))) ; avoid walls and water

(define (translate-cell c)
  (case c
    ((#\#) (make-wall))
    ((#\~) (make-water))
    ((#\.) (make-empty))
    ((#\@) (make-empty))
    (else  (error "illegal cell" c))))

; **** STATE ****

(define trace? #f)
(define visualize? #f)

(define width+2 #f)      ; size of the board
(define height+2 #f)

(define N #f)
(define S #f)
(define E 1)
(define W -1)

(define nb-steps 1)

(define board #f)        ; the board (as a 1D vector)

(define bases '())       ; location of the bases
(define rid 0)           ; our robot id
(define capacity 0)      ; our capacity
(define money 0)         ; our money
(define robot-ids '())   ; the list of robot ids
(define robots '())      ; knowledge of the robots
(define package-ids '()) ; the list of package ids
(define packages '())    ; knowledge of the packages

(define me #f)           ; this robot

(define (col-row->pos col row)
  (+ (* row width+2) col))

(define (pos->row x)
  (quotient x width+2))

(define (pos->col x)
  (modulo x width+2))

(define (visu:water? col row)
  (water? (col-row->pos col row)))

(define (visu:wall? col row)
  (wall? (col-row->pos col row)))

(define (visu:base? col row)
  (let ((pos (col-row->pos col row)))
    (memq pos bases)))

(define (visu:robot-id col row)
  (let ((pos (col-row->pos col row)))
    (let ((r (pos->robot pos)))
      (if r
          (robot-id r)
          #f))))

(define (visu:robot-self? id)
  (= id rid))

(define (visu:package-likelihood col row)
  (let ((pos (col-row->pos col row)))
    (let ((x (pos->packages pos)))
      (if (uncertain? x)
          (if (null? (->certain x))
              0.5
              1.0)
          (if (null? x)
              0.0
              1.0)))))

(define (visu:robot-package-likelihood id)
  (let ((r (id->robot id)))
    (let ((x (robot-packages r)))
      (if (uncertain? x)
          (if (null? (->certain x))
              0.5
              1.0)
          (if (null? x)
              0.0
              1.0)))))

(define (visu:message)
  (with-output-to-string
    (lambda ()
      (pp (list (list 'width: (- width+2 2) 'height: (- height+2 2))
                (list 'money: money)
                (list 'cpu-time-per-step:
                      (let ()
                        (declare (flonum))
                        (/ (cpu-time)
                           (exact->inexact nb-steps))))
;                'robots: robots
                )))))

(define (visu:info col row)
'
  (let ((pos (col-row->pos col row)))
    (number->string (distance-to pos distance-map-from-me)))
'
  (let ((pos (col-row->pos col row)))
    (let ((x (control pos control-map-excluding-me)))
      (if (pair? x)
;          (number->string (length (cdr x)))
          (if (or (null? (cdr x)) (> (length (cdr x)) 1))
              "??"
              (number->string (cadr x)))
          #f)))
#f
)

(define (visu:dot? col row)
  (let ((pos (col-row->pos col row)))
    (let loop ((lst (robot-packages me)))
      (if (pair? lst)
          (let ((p (id->package (car lst))))
            (if (= pos (package-pos p))
                #t
                (loop (cdr lst))))
          #f))))

(define distance-map-from-me #f)
(define control-map #f)
(define control-map-excluding-me #f)
(define RO-path #f)

(define distance-map-cache (##make-table))

(define (compute-distance-map pos)
  (let ((dm
         (begin
           '
           (or (##table-ref distance-map-cache pos #f)
               (let ((dm (actually-compute-distance-map pos)))
                 (##table-set! distance-map-cache pos dm)
                 dm))
           (actually-compute-distance-map pos))))
    dm))

(define-learnable-parameter fancy-dist-map 0 1)
(define fancy-dist-map-aux (= (fancy-dist-map) 1))

(define (actually-compute-distance-map pos)
  (if fancy-dist-map-aux
      (compute-distance-map-with-cost pos cost-function)
      (compute-distance-map-with-unit-cost pos)))

(define infinite 9999999)

(define-learnable-parameter enemy-cost 0 3)
(define enemy-cost-aux (* 5 (+ (enemy-cost) 1)))

(define-learnable-parameter drowning-cost 0 7)
(define drowning-cost-aux (* 10 (+ (drowning-cost) 1)))

(define-learnable-parameter unvisited-cost 0 3)
(define unvisited-cost-aux (+ (* 2 (unvisited-cost)) 1))

(define-learnable-parameter empty-cost 0 3)
(define empty-cost-aux (+ (empty-cost) 1))

(define (cost-function pos) ; must return #f or an integer > 0
  (cond ((obstacle? pos)
         #f)
        ((pos->robot pos)
         enemy-cost-aux)
        ((next-to-water-and-pushable? pos rid)
         drowning-cost-aux) ; don't go close to the water if someone might push you
        (else
         (let ((p (pos->packages pos)))
           (if (uncertain? p)
               unvisited-cost-aux ; good to visit unvisited tiles
               (if (null? p)
                   empty-cost-aux ; try not to visit tiles known to be empty
                   1)))))) ; it is good to visit a non-empty tile

(define (next-to-water? pos)

  (define-macro (check dir)
    `(water? (+ pos ,dir)))

  (or (check N)
      (check S)
      (check E)
      (check W)))

(define (next-to-water-and-pushable? pos rob-id)

  (define-macro (check dir)
    `(and (water? (+ pos ,dir))
          (not (obstacle? (- pos ,dir)))
          (let ((c (control (- pos ,dir) control-map)))
            (not (memq rob-id (cdr c))))))

  (or (check N)
      (check S)
      (check E)
      (check W)))

(define (compute-distance-map-with-cost pos cost-function)

  (define 2^20 1048576)
  (define -2^20 -1048576)
  (define 2^20-1 1048575)

  (let* ((n (vector-length board))
         (dm (make-vector n 0)))

    (let loop1 ((i (- n 1)))
      (if (not (< i 0))
          (let ((x (cost-function i)))
            (if x
                (vector-set! dm i (- x))
                (vector-set! dm i (+ infinite 1)))
            (loop1 (- i 1)))))

    (vector-set! dm 0 pos)
    (vector-set! dm pos 1)

    (let loop2 ((d 1) (step 1))
      (let ((head (vector-ref dm 0)))
        (if (not (= head 1))
            (begin
              (vector-set! dm 0 1)
              (let loop3 ((p head)
                          (min-remaining 999)) ; anything above 511 is OK
                (if (= p 1)
                    (loop2 (+ d step)
                           min-remaining)
                    (let* ((next
                            (vector-ref dm p))
                           (remaining
                            (- (##fixnum.arithmetic-shift-right next 20)
                               step)))
                      (if (> remaining 0) ; this tile will be processed later
                          (let ((new
                                 (+ (##fixnum.arithmetic-shift-left remaining 20)
                                    (vector-ref dm 0))))
                            (vector-set! dm p new)
                            (vector-set! dm 0 p)
                            (if (< remaining min-remaining)
                                (loop3 (bitwise-and next 2^20-1)
                                       remaining)
                                (loop3 (bitwise-and next 2^20-1)
                                       min-remaining)))
                          (let ()

                            (define-macro (visit position)
                              `(let* ((i ,position)
                                      (x (vector-ref dm i)))
                                 (if (< x 0) ; not yet visited?
                                     (let ((remaining (- x)))
                                       (vector-set! dm i
                                         (bitwise-ior
                                          (vector-ref dm 0)
                                          (##fixnum.arithmetic-shift-left
                                           remaining
                                           20)))
                                       (vector-set! dm 0 i)
                                       remaining)
                                     999)))

                            (vector-set! dm p d)

                            (loop3 (bitwise-and next 2^20-1)
                                   (min (visit (+ p width+2))
                                        (min (visit (- p width+2))
                                             (min (visit (+ p 1))
                                                  (min (visit (- p 1))
                                                       min-remaining))))))))))))))

    dm))

(define (compute-distance-map-with-unit-cost pos)
  (let* ((n (vector-length board))
         (dm (make-vector n 0)))

    (let loop1 ((i (- n 1)))
      (if (not (< i 0))
          (begin
            (cond ((obstacle? i)
                   (vector-set! dm i (+ infinite 1)))
                  (else
                   (vector-set! dm i 0)))
            (loop1 (- i 1)))))

    (vector-set! dm 0 pos)
    (vector-set! dm pos 1)

    (let loop2 ((d 1))
      (let ((head (vector-ref dm 0)))
        (if (not (= head 1))
            (begin
              (vector-set! dm 0 1)
              (let loop3 ((p head))
                (if (= p 1)
                    (loop2 (+ d 1))
                    (let ()

                      (define-macro (visit position)
                        `(let* ((i ,position)
                                (x (vector-ref dm i)))
                           (if (= x 0) ; not yet visited?
                               (begin
                                 (vector-set! dm i (vector-ref dm 0))
                                 (vector-set! dm 0 i)))))

                      (visit (+ p width+2))
                      (visit (- p width+2))
                      (visit (+ p 1))
                      (visit (- p 1))

                      (let ((next (vector-ref dm p)))
                        (vector-set! dm p d)
                        (loop3 next)))))))))

    dm))

(define (distance-to pos dm)
  (if (or (< pos 0)
          (not (< pos (vector-length dm))))
      infinite
      (let ((d (vector-ref dm pos)))
        (if (<= d 0)
            infinite
            (- d 1)))))

(define (compute-control-map! rob-ids)
  (let* ((n (vector-length board))
         (cm (make-vector n #f))
         (dm (make-vector n 0)))

    (let loop0 ((i (- n 1)))
      (if (not (< i 0))
          (begin
            (cond ((obstacle? i)
                   (vector-set! cm i '(9999999))
                   (vector-set! dm i (+ infinite 1)))
                  (else
                   (vector-set! dm i 0)))
            (loop0 (- i 1)))))

    (let ((starts
           (let loop1 ((lst1 rob-ids)
                       (lst2 '()))
             (if (pair? lst1)
                 (let* ((id (car lst1))
                        (pos (robot-pos (id->robot id))))
                   (if (= (vector-ref dm pos) 0)
                       (begin
                         (vector-set! cm pos (list 0 id))
                         (vector-set! dm pos 1)
                         (loop1 (cdr lst1)
                                (cons (cons id pos) lst2)))
                       (loop1 (cdr lst1)
                              lst2)))
                 lst2))))

      (let loop2 ((d 1) (starts starts))
        (let loop3 ((lst1 starts)
                    (lst2 '()))
          (if (null? lst1)
              (if (null? lst2)
                  cm
                  (loop2 (+ d 1) lst2))
              (let* ((x (car lst1))
                     (id (car x))
                     (head (cdr x)))
                (vector-set! dm 0 1)
                (let loop4 ((p head))
                  (if (= p 1)
                      (let ((new-head (vector-ref dm 0)))
                        (if (= new-head 1)
                            (loop3 (cdr lst1)
                                   lst2)
                            (loop3 (cdr lst1)
                                   (cons (cons id new-head) lst2))))
                      (let ()

                        (define-macro (visit position)
                          `(let* ((i ,position)
                                  (x (vector-ref cm i)))
                             (if (pair? x) ; it has been visited
                                 (if (= (car x) d) ; but in this round
                                     (let ((rest (cdr x)))
                                       (if (not (= (car rest) id))
                                           (vector-set! cm i (cons d (cons id rest))))))
                                 (begin ; it was not yet visited
                                   (vector-set! cm i (list d id))
                                   (vector-set! dm i (vector-ref dm 0))
                                   (vector-set! dm 0 i)))))

                        (visit (+ p width+2))
                        (visit (- p width+2))
                        (visit (+ p 1))
                        (visit (- p 1))

                        (let ((next (vector-ref dm p)))
                          (vector-set! dm p d)
                          (loop4 next))))))))))))

(define (control pos control-map)
  (if (or (< pos 0)
          (not (< pos (vector-length control-map))))
      '(9999999)
      (let ((x (vector-ref control-map pos)))
        (or x
            '(9999999)))))

(define (pay bid port)
  (set! money (- money (abs bid))))
;  (if (< money 0)
;    (begin
;      (close-port port)
;      (exit 0))));;;;;;;;;;;;;;;;;;;;;

;------------------------------------------------------------------------------

(define-type action
  extender: define-type-of-action
  (voice)
  (bid)
)

(define-type-of-action Move
  (direction)
)

(define-type-of-action Pick
  (packages)
)

(define-type-of-action Drop
  (packages)
)

(define (copy-action act)
  (cond ((Move? act)
         (make-Move (action-voice act)
                    (action-bid act)
                    (Move-direction act)))
        ((Pick? act)
         (make-Pick (action-voice act)
                    (action-bid act)
                    (Pick-packages act)))
        ((Drop? act)
         (make-Drop (action-voice act)
                    (action-bid act)
                    (Drop-packages act)))
        (else
         (error "unknown action"))))

(define (action=? act1 act2)
  (cond ((Move? act1)
         (and (Move? act2)
              (eq? (Move-direction act1)
                   (Move-direction act2))))
        ((Pick? act1)
         (and (Pick? act2)
              (equal? (sort-list (Pick-packages act1) <)
                      (sort-list (Pick-packages act2) <))))
        ((Drop? act1)
         (and (Drop? act2)
              (equal? (sort-list (Drop-packages act1) <)
                      (sort-list (Drop-packages act2) <))))
        (else
         #f)))

(define (action<? act1 act2)
  (cond ((Move? act1)
         (cond ((Move? act2)
                (string<? (symbol->string (Move-direction act1))
                          (symbol->string (Move-direction act2))))
               (else
                #t)))
        ((Pick? act1)
         (cond ((Move? act2)
                #f)
               ((Pick? act2)
                (int-list<? (sort-list (Pick-packages act1) <)
                            (sort-list (Pick-packages act2) <)))
               (else
                #t)))
        ((Drop? act1)
         (cond ((Drop? act2)
                (int-list<? (sort-list (Drop-packages act1) <)
                            (sort-list (Drop-packages act2) <)))
               (else
                #f)))))

(define (int-list<? lst1 lst2)
  (cond ((null? lst1)
         (cond ((null? lst2)
                #f)
               (else
                #t)))
        (else
         (cond ((null? lst2)
                #f)
               (else
                (let ((x1 (car lst1))
                      (x2 (car lst2)))
                  (cond ((< x1 x2)
                         #t)
                        ((> x1 x2)
                         #f)
                        (else
                         (int-list<? (cdr lst1) (cdr lst2))))))))))

(define (combine-actions actions)

  ; don't ignore this "intelligent" stuff

  (let loop ((lst1 (sort-list actions action<?))
             (lst2 '()))
    (if (pair? lst1)
        (let ((act1 (car lst1))
              (rest (cdr lst1)))
          (if (pair? rest)
              (let ((act2 (car rest)))
                (if (action=? act1 act2)
                    (let ((new (copy-action act1)))
                      (action-bid-set! new
				       (max (action-bid act1)
                                            (action-bid act2)))
                      (action-voice-set! new
                                         (+ (action-voice act1)
                                            (action-voice act2)))
                      (loop (cons new
                                  (cdr rest))
                            lst2))
                    (loop rest
                          (cons act1 lst2))))
              (loop rest
                    (cons act1 lst2))))
        (sort-list lst2
                   (lambda (x y) (> (action-voice x) (action-voice y))))))

;;;;;;;  actions;;;;;;;;;don't just return the list unchanged
)

(define (write-action act port)
  (display (action-bid act) port)
  (display " " port)
  (cond ((Move? act)
         (display "Move " port)
         (display (Move-direction act) port))
        ((Pick? act)
         (display "Pick" port)
         (for-each (lambda (x) (display " " port) (display x port))
                   (Pick-packages act)))
        ((Drop? act)
         (display "Drop" port)
         (for-each (lambda (x) (display " " port) (display x port))
                   (Drop-packages act)))
        (else
         (error "unknown action")))
  (newline port)
  (force-output port))

;------------------------------------------------------------------------------

(##include "rand.scm")

;------------------------------------------------------------------------------

(define (remq x lst)
  (cond ((null? lst)
         '())
        ((eq? (car lst) x)
         (cdr lst))
        (else
         (cons (car lst) (remq x (cdr lst))))))

(define (sort-list l <?)

  (define (mergesort l)

    (define (merge l1 l2)
      (cond ((null? l1) l2)
            ((null? l2) l1)
            (else
             (let ((e1 (car l1)) (e2 (car l2)))
               (if (<? e1 e2)
                 (cons e1 (merge (cdr l1) l2))
                 (cons e2 (merge l1 (cdr l2))))))))

    (define (split l)
      (if (or (null? l) (null? (cdr l)))
        l
        (cons (car l) (split (cddr l)))))

    (if (or (null? l) (null? (cdr l)))
      l
      (let* ((l1 (mergesort (split l)))
             (l2 (mergesort (split (cdr l)))))
        (merge l1 l2))))

  (mergesort l))

(define (keep keep? lst)
  (cond ((null? lst)       '())
        ((keep? (car lst)) (cons (car lst) (keep keep? (cdr lst))))
        (else              (keep keep? (cdr lst)))))

(define (pos-in-list x l)
  (let loop ((l l) (i 0))
    (cond ((not (pair? l)) #f)
          ((eq? (car l) x) i)
          (else            (loop (cdr l) (+ i 1))))))

(define (make-table)
  (make-stretchable-vector #f))

(define (table-ref t i)
  (stretchable-vector-ref t i))

(define (table-set! t i val)
  (stretchable-vector-set! t i val))

(define (make-stretchable-vector init)
  (vector '#() init))

(define (stretchable-vector-length sv)
  (vector-length (vector-ref sv 0)))

(define (stretchable-vector-ref sv i)
  (let* ((v (vector-ref sv 0))
         (len (vector-length v)))
    (if (< i len)
      (vector-ref v i)
      (vector-ref sv 1))))

(define (stretchable-vector-set! sv i x)
  (let* ((v (vector-ref sv 0))
         (len (vector-length v)))
    (if (< i len)
      (vector-set! v i x)
      (let ((new-v
              (stretch-vector
                v
                (+ (max i (quotient (* len 3) 2)) 1) ; make 50% bigger at least
                (vector-ref sv 1))))
        (vector-set! sv 0 new-v)
        (vector-set! new-v i x)))))

(define (stretch-vector v n init)
  (let ((len (vector-length v))
        (new-v (make-vector n)))
    (let loop1 ((i 0))
      (if (< i len)
        (begin
          (vector-set! new-v i (vector-ref v i))
          (loop1 (+ i 1)))))
    (let loop2 ((i len))
      (if (< i n)
        (begin
          (vector-set! new-v i init)
          (loop2 (+ i 1)))))
    new-v))

(define (stretchable-vector-copy sv)
  (let* ((v1 (vector-ref sv 0))
         (n (vector-length v1))
         (v2 (make-vector n)))
    (let loop ((i (- n 1)))
      (if (>= i 0)
        (begin
          (vector-set! v2 i (vector-ref v1 i))
          (loop (- i 1)))
        (vector v2 (vector-ref sv 1))))))

(define (stretchable-vector-for-each proc sv)
  (let* ((v (vector-ref sv 0))
         (n (vector-length v)))
    (let loop ((i 0))
      (if (< i n)
        (begin
          (proc (vector-ref v i) i)
          (loop (+ i 1)))))))

;------------------------------------------------------------------------------

(define (bases-reachable-from pos)
  (let ((dm (compute-distance-map pos)))
    (keep (lambda (base-pos)
            (< (distance-to base-pos dm) infinite))
          bases)))

(define base-distance-map #f)

(define (compute-base-distance-map)
  (let ((n (length bases)))
    (let ((bdm (make-vector n '#())))
      (if (= n 0)
          (begin
            (set! base-distance-map bdm))
          (let loop1 ((lst1 (cdr bases))
                      (i 1))
            (if (pair? lst1)
                (let ((base-pos1 (car lst1))
                      (v (make-vector i #f)))
                  (vector-set! bdm i v)
                  (let ((dm (compute-distance-map base-pos1)))
                    (let loop2 ((lst2 bases)
                                (j 0))
                      (if (< j i)
                          (let ((base-pos2 (car lst2)))
                            (let ((d (distance-to base-pos2 dm)))
                              (if (< d infinite)
                                  (vector-set! v j d))
                              (loop2 (cdr lst2) (+ j 1))))
                          (loop1 (cdr lst1) (+ i 1))))))
                (begin
                  (set! base-distance-map bdm))))))))

(define (base-distance-between base-pos1 base-pos2)
  (base-distance-between-aux
   (pos-in-list base-pos1 bases)
   (pos-in-list base-pos2 bases)))

(define (base-distance-between-aux base-id1 base-id2)
  (if (= base-id1 base-id2)
      0
      (if (< base-id1 base-id2)
          (vector-ref (vector-ref base-distance-map base-id2) base-id1)
          (vector-ref (vector-ref base-distance-map base-id1) base-id2))))

(define (uncertain-or-non-empty-bases)
  (keep (lambda (pos)
          (let ((x (pos->packages pos)))
            (or (uncertain? x)
                (not (null? x)))))
        bases))

(define (uncertain-bases)
  (keep (lambda (pos)
          (let ((x (pos->packages pos)))
            (uncertain? x)))
        bases))

;------------------------------------------------------------------------------

(define (closest pos-list)
  (if (null? pos-list)
      #f
      (cdr (choose-lowest
            (map
             (lambda (pos)
               (let ((dist (distance-to pos distance-map-from-me)))
                 (cons dist pos)))
             pos-list)))))

;------------------------------------------------------------------------------

(define (choose-highest lst)
  (cdr (car (sort-list (map (lambda (x)
                              (cons (+ (* 100 (car x)) (rand 100))
                                    x))
                            lst)
                       (lambda (x y) (> (car x) (car y)))))))

(define (choose-lowest lst)
  (cdr (car (sort-list (map (lambda (x)
                              (cons (+ (* 100 (car x)) (rand 100))
                                    x))
                            lst)
                       (lambda (x y) (< (car x) (car y)))))))

;------------------------------------------------------------------------------

(define (possible-directions pos)
  (keep (lambda (dir)
          (not (obstacle? (displace pos dir))))
        '(N E S W)))

;------------------------------------------------------------------------------

(define (scan pos horizon init combine)
  (let loop1 ((dy (- horizon)) (result init))
    (if (<= dy horizon)
        (let ((n (- horizon (abs dy))))
          (let loop2 ((dx (- n)) (result result))
            (if (<= dx n)
                (loop2 (+ dx 1)
                       (combine (+ pos (+ (* dy width+2) dx)) result))
                (loop1 (+ dy 1) result))))
        result)))

;------------------------------------------------------------------------------

(define (direction-to pos)
  (let* ((my-pos (robot-pos me))
         (dm (compute-distance-map pos)))

    (define (consider-direction dir)
      (let ((dist (distance-to (displace my-pos dir) dm)))
        (cons dist dir)))

    (let ((dist-dirs
           (map consider-direction (possible-directions my-pos))))
      (if (null? dist-dirs)
          #f
          (cdr (choose-lowest dist-dirs))))))

;------------------------------------------------------------------------------

(define-macro (desire-nothing)
  `(begin
     (if trace?
         (pp (list expert-name 'desires 'nothing)))
     '()))

(define-macro (desire-urgent-Move dir voice)
  `(begin
     (if trace?
         (pp (list expert-name 'desires 'urgent 'Move ,dir)))
     (list (make-Move ,voice 25 ,dir))))

(define-macro (desire-Move dir voice)
  `(begin
     (if trace?
         (pp (list expert-name 'desires 'Move ,dir)))
     (list (make-Move ,voice 1 ,dir))))

(define-macro (desire-Pick lst voice)
  `(begin
     (if trace?
         (pp (list expert-name 'desires 'Pick ,lst)))
     (list (make-Pick ,voice 1 ,lst))))

(define-macro (desire-Drop lst voice)
  `(begin
     (if trace?
         (pp (list expert-name 'desires 'Drop ,lst)))
     (list (make-Drop ,voice 1 ,lst))))

;------------------------------------------------------------------------------

(define foldr
  (lambda (f base lst)
    (if (null? lst)
        base
        (f (car lst) (foldr f base (cdr lst))))))

(define foldl
  (lambda (f base lst)
    (if (null? lst)
        base
        (foldl f (f base (car lst)) (cdr lst)))))

(define sum
  (lambda (lst)
    (foldl + 0 lst)))

(define product
  (lambda (lst)
    (foldl * 1 lst)))

;------------------------------------------------------------------------------

(define (packages-known-to-be-pickable pos)
  (let ((packages (->certain (pos->packages pos))))
    (if (null? packages) ; there are no packages here!

        '()

        (let* ((current-load
                (sum (map (lambda (id)
                            (package-weight (id->package id)))
                          (robot-packages me))))
               (largest-to-smallest
                (sort-list
                 (map id->package packages)
                 (lambda (x y) (> (package-weight x) (package-weight y))))))
          (let loop ((lst largest-to-smallest)
                     (pickable '())
                     (free-capacity (- capacity current-load)))
            (if (pair? lst)
                (let* ((p (car lst))
                       (weight (package-weight p)))
                  (if (and (< (distance-to (package-pos p)
                                           distance-map-from-me)
                              infinite)
                           (<= weight free-capacity))
                      (loop (cdr lst)
                            (cons (package-id p) pickable)
                            (- free-capacity weight))
                      (loop (cdr lst)
                            pickable
                            free-capacity)))
                pickable))))))

;------------------------------------------------------------------------------

(define packages-brought-to-dest
  (lambda ()
    (let* ((pos (robot-pos me))
	   (p-ids (robot-packages me))
	   (br? (lambda (p-id)
		  (equal? (package-pos (id->package p-id)) pos))))
      (filter br? p-ids))))

;------------------------------------------------------------------------------

(define RO-trace
  (lambda ()
    (letrec
	((p2p
	  (lambda (pos)
	    (list (pos->col pos) (pos->row pos))))
	 (interesting-positions
	  (keep (lambda (bpos)
		  (let ((packs (pos->packages bpos)))
		    (if (uncertain? packs)
			#t
			(pair? packs))))
		bases)))
      (display "********** RO trace **********\n")
      (display "Robot at ")
      (pp (p2p (robot-pos me)))
      (display "Packages on its back:\n")
      (pp (map (lambda (p-id) (p2p (package-pos (id->package p-id))))
	       (robot-packages me)))
      (display "Pick possibilities:\n")
      (pp (map p2p interesting-positions))
      (display "RO path:\n")
      (pp (map p2p RO-path))
      (display "********** RO trace **********\n"))))

;------------------------------------------------------------------------------

; Fonction de nettoyage du chemin retourne par `visit-and-deliver'
; Enleve la position courante du debut du chemin

(define cleanup-RO-path
  (lambda (path)
    (let ((here (robot-pos me)))
      (let loop ((path path))
	(if (null? path)
	    path
	    (let ((hd (car path)))
	      (if (equal? hd here)
		  (loop (cdr path))
		  path)))))))

;------------------------------------------------------------------------------

; Traceur de chemins de type "visit-and-deliver"

; Ramasse les infos pour vnd-solve et les lui passe

; Note: il y un mensonge qui est conte a vnd-solve; c'est le poids
; moyen des paquets connus a date qui sert d'estimer du load total par
; home base.  vnd-solve s'en sacre completement.

(define filter
  (lambda (pred? l)
    (if (null? l)
	'()
	(let ((hd (car l)))
	  (if (pred? hd)
	      (cons hd (filter pred? (cdr l)))
	      (filter pred? (cdr l)))))))

; Bornes incluses
(define list-interval
  (lambda (a b)
    (if (> a b)
	'()
	(cons a (list-interval (+ a 1) b)))))

(define visit-and-deliver
  (lambda ()
    (let* ((bases2visit (vnd-find-bases))  ; Cases qui peuvent contenir un paq.
	   (vnd-bs (list-interval 1 (length bases2visit)))
	   (my-packs (robot-packages me))               ; Destinations et poids
	   (op (lambda (pid)
		 (let ((pack (id->package pid)))
		   (list (package-pos pack) (package-weight pack)))))
	   (dws (map op my-packs))
	   (dw<? (lambda (dw1 dw2) (< (car dw1) (car dw2))))
	   (dws (sort-list dws dw<?))
	   (dws (let loop ((dws dws)) ; no dup
		  (cond ((null? dws)
			 '())
			((null? (cdr dws))
			 dws)
			(else
			 (let* ((dw1 (car dws))
				(dw2 (cadr dws)))
			   (if (equal? (car dw1) (car dw2))
			       (let ((dw0 (list (car dw1)
						(+ (cadr dw1) (cadr dw2)))))
				 (loop (cons dw0 (cddr dws))))
			       (cons dw1 (loop (cdr dws)))))))))
	   (vnd-ds (list-interval (- (length dws)) -1))
	   (my-pos (robot-pos me))                                 ; Traducteur
	   (trans-base-v (list->vector bases2visit))
	   (trans-base
	    (lambda (b)
	      (vector-ref trans-base-v (- b 1))))
	   (trans-dest-v (list->vector (map car dws)))
	   (trans-dest
	    (lambda (d)
	      (vector-ref trans-dest-v (- -1 d))))
	   (trans
	    (lambda (n)
	      (if (> n 0)
		  (trans-base n)
		  (if (< n 0)
		      (trans-dest n)
		      my-pos))))
	   (dist-base-maps-v                                 ; Fonction de cout
	    (list->vector
	     (map compute-distance-map-with-unit-cost bases2visit)))
	   (dist-dest-maps-v
	    (list->vector
	     (map compute-distance-map-with-unit-cost (map car dws))))
	   (dist-self-map
	    (compute-distance-map-with-unit-cost my-pos))
	   (vnd-cost-f
	    (lambda (n1 n2)
	      (cond ((> n1 0)
		     (let ((dist-map (vector-ref dist-base-maps-v (- n1 1))))
		       (distance-to (trans n2) dist-map)))
		    ((< n1 0)
		     (let ((dist-map (vector-ref dist-dest-maps-v (- -1 n1))))
		       (distance-to (trans n2) dist-map)))
		    (else
		     (distance-to (trans n2) dist-self-map)))))
	   (vnd-cap capacity)                                        ; Capacite
	   (vnd-avg-base-load                         ; Poids moyen des paquets
	    (let loop ((pids package-ids) (nb 0) (total-w 0))
	      (if (null? pids)
		  (if (zero? nb)
		      (max (quotient capacity 2) 1)
		      (max (quotient total-w nb) 1))
		  (let* ((pid (car pids))
			 (pack (id->package pid)))
		    (if pack
			(loop (cdr pids)
			      (+ nb 1)
			      (+ total-w (package-weight pack)))
			(loop (cdr pids) nb total-w))))))
	   (dest-load-v (list->vector (map cadr dws))) ; Fonction de load/dest.
	   (vnd-load-f
	    (lambda (d)
	      (vector-ref dest-load-v (- -1 d))))
	   (path (vnd-solve vnd-bs                         ; Calcul d'un chemin
			    vnd-ds
			    vnd-cost-f
			    vnd-cap
			    vnd-avg-base-load
			    vnd-load-f)))
      (map trans path))))

(define vnd-find-bases
  (lambda ()
    (let* ((all-bases bases)             ; Bases qui peuvent contenir un paquet
	   (interesting?
	    (lambda (bpos)
	      (let ((packs (pos->packages bpos)))
		(if (uncertain? packs)
		    #t
		    (pair? packs)))))
	   (bs (filter interesting? all-bases)))
      (if (not (null? bs))
	  bs
	  (let loop ((y (- height+2 2))
		     (x (- width+2 2))
		     (pos (col-row->pos (- width+2 2) (- height+2 2)))
		     (acc '())
		     (any? #f))
	    (cond ((< y 1)
		   (if (not (null? acc))
		       acc
		       (loop (- height+2 2)
			     (- width+2 2)
			     (col-row->pos (- width+2 2) (- height+2 2))
			     '()
			     #t)))
		  ((< x 1)
		   (loop (- y 1)
			 (- width+2 2)
			 (col-row->pos (- width+2 2) (- y 1))
			 acc
			 any?))
		  (else
		   (let ((p-ids (pos->packages pos)))
		     (if (uncertain? p-ids)
			 (if any?
			     (loop y (- x 1) (+ pos W) (cons pos acc) #t)
			     (loop y (- x 1) (+ pos W) acc #f))
			 (if (null? p-ids)
			     (loop y (- x 1) (+ pos W) acc any?)
			     (loop y
				   (- x 1)
				   (+ pos W)
				   (cons pos acc)
				   any?)))))))))))

; Implantation d'une heuristique pour tracer des routes
; visit-and-deliver

; Ce sont les home bases non-visitees qu'il faut visiter.  Les points
; de destination sont ceux des paquets qui nous appartiennent.  Un
; estime de la charge rajoutee a chaque home base est pris en compte.
; Le chemin tente d'eviter de surcharger le robot.  En ce sens, les
; destinations sont privilegiees sur les home bases.

; Recoit:
; - Une liste de home bases qui restent a visiter.  Liste de numeros.
; - Une liste de destinations.  Liste de numeros.  Les numeros
;   negatifs sont les destinations; 0 est la position actuelle; les
;   numeros positifs sont les home bases.
; - Une fonction de cout:: numero , numero -> cout.
; - Capacite du robot.
; - Poids moyen des paquets a ramasser par home base (devrait avoir un
;   defaut egal a capacite/2).
; - Charge a deposer par destination:: numero -> poids.  Le poids
;   devrait etre > 0 pour toutes les destinations.
; Retourne:
; - Un chemin partant de la position actuelle et qui passe par toutes
;   les home bases et les destinations.  Liste de numeros.

; Algo:
; - Insere d'abord toutes les destinations de la plus "lourde" a la
;   plus "legere".
; - Insere ensuite dans l'ordre de la plus petite penalite a la plus
;   grande les home bases A CONDITION que l'insertion la plus
;   profitable ne se fasse pas apres la derniere destination.  Une
;   insertion est interdite si elle cause le depassement probable de
;   la capacite
; - Pour toutes les home bases non-inserees, s'il en reste, deleguer a
;   un TSP qui part de la derniere destination.

(define vnd-solve
  (lambda (bs ds cost-f cap avg-base-load load-f)
    (letrec
	((load0
	  (apply + (map load-f ds)))
	 (infinity
	  (max (+ cap 1) 200000000 (+ avg-base-load 1)))
	 (sort-ds
	  (lambda (ds)                ; -> ds (en ordre de charge decroissante)
	    (let ((rel<= (lambda (d1 d2) (>= (load-f d1) (load-f d2)))))
	      (sort-list ds rel<=))))
	 (insert-1-dest
	  (lambda (path d)                                            ; -> path
	    (let loop ((path path) (best infinity))
	      (if (null? (cdr path))
		  (let* ((d0 (car path))
			 (c (cost-f d0 d)))            ; cout insertion d0-d-()
		    (if (< c best)
			(list d0 d)
			#f))
		  (let* ((d0 (car path))
			 (rst (cdr path))
			 (d1 (car rst))
			 (c (- (+ (cost-f d0 d)        ; cout insertion d0-d-d1
				  (cost-f d d1))
			       (cost-f d0 d1)))
			 (res (loop rst (min best c))))
		    (if res
			(cons d0 res)                         ; meilleur + loin
			(if (< c best)
			    (cons d0 (cons d rst))               ; meilleur ici
			    #f)))))))                          ; meilleur avant
	 (insert-dests
	  (lambda (ds path)                                           ; -> path
	    (let ((ds (sort-ds ds)))
	      (foldl insert-1-dest path ds))))
	 (i-cost-f
	  (lambda (b path load)                                       ; -> cost
	    (let* ((avg avg-base-load)
		   (d0 (car path))
		   (rst (cdr path)))
	      (if (null? rst)
		  (cost-f d0 b)
		  (let* ((d1 (car rst))
			 (det (- (+ (cost-f d0 b)              ; cout du detour
				    (cost-f b d1))
				 (cost-f d0 d1))))
		    (let loop ((load (+ load avg)) (path rst))   ; ins. legale?
		      (if (null? path)
			  det
			  (let* ((d0 (car path))
				 (rst (cdr path)))
			    (if (< d0 0)
				(loop (- load (load-f d0)) rst)
				(let ((load2 (+ load avg)))
				  (if (> load2 cap)
				      infinity
				      (loop load2 rst))))))))))))
	 (list-insert
	  (lambda (l i b)
	    (cond ((zero? i)
		   (cons b l))
		  ((null? l)
		   (error "Invalid index in list-insert in vnd-solve."))
		  (else
		   (cons (car l) (list-insert (cdr l) (- i 1) b))))))
	 (insert-1-base-try
	  (lambda (b path-all)                           ; -> (cost path) ou #f
	    (let loop ((path path-all)
		       (load load0)
		       (pos 1)
		       (i-pos #f)
		       (best infinity))
	      (let ((rst (cdr path)))
		(if (null? rst)
		    (let ((c (i-cost-f b path load)))           ; cout ins. fin
		      (if (< c best)
			  #f                          ; ins. a la fin interdite
			  (list best (list-insert path-all i-pos b))))
		    (let* ((c (i-cost-f b path load))
			   (load2 (min (+ load avg-base-load) cap)))
		      (if (< c best)
			  (loop (cdr path) load2 (+ pos 1) pos c)
			  (loop (cdr path) load2 (+ pos 1) i-pos best))))))))
	 (insert-base-try
	  (lambda (bs path)                                ; -> (bs path) ou #f
	    (let loop ((pbs '()) (fbs bs) (best infinity))
	      (if (null? fbs)
		  #f
		  (let* ((b (car fbs))
			 (res (insert-1-base-try b path))
			 (c (if res (car res) infinity))
			 (path1 (if res (cadr res) #f))
			 (res (loop (cons b pbs) (cdr fbs) (min best c))))
		    (if res
			res
			(if (< c best)
			    (list (append pbs (cdr fbs)) path1)
			    #f)))))))
	 (insert-bases-1
	  (lambda (bs path)                                      ; -> (path bs)
	    (let loop ((bs bs) (path path))
	      (if (null? bs)
		  (list path bs)                      ; plus de bases a inserer
		  (let ((res (insert-base-try bs path)))
		    (if res
			(loop (car res) (cadr res))         ; insertion reussie
			(list path bs)))))))            ; insertion ratee: fini
	 (last
	  (lambda (l)                                                  ; -> obj
	    (if (null? l)
		(error "Empty list in last.")
		(let loop ((l l))
		  (if (null? (cdr l))
		      (car l)
		      (loop (cdr l)))))))
	 (insert-bases-2
	  (lambda (bs path)                                           ; -> path
	    (let* ((start-d (last path))
		   (trans-v (list->vector (cons start-d bs)))
		   (trans (lambda (n) (vector-ref trans-v n)))
		   (tsp-n (length bs))
		   (tsp-cost (lambda (n1 n2) (cost-f (trans n1) (trans n2))))
		   (res (tsp-solve tsp-n tsp-cost))
		   (path2 (map trans res)))
	      (append path (cdr path2))))))
      (let* ((path0
	      (list 0))
	     (path1
	      (insert-dests ds path0))
	     (res
	      (insert-bases-1 bs path1))
	     (path2
	      (car res))
	     (rem-bs
	      (cadr res)))
	(insert-bases-2 rem-bs path2)))))

; Petit test
(define vnd-ex-bs '(1 2 3 4))

(define vnd-ex-ds '(-1 -2))

(define vnd-ex-cost-f
  (lambda (x y)
    (vector-ref
     (vector-ref
      ;    -2 -1  0  1  2  3  4
      '#(#( 0 11  8  2  6 20 10)   ; -2
	 #(11  0  8 12 12 16  9)   ; -1
	 #( 8  8  0  7  3 12  3)   ;  0
	 #( 2 12  7  0  5 19  9)   ;  1
	 #( 6 12  3  5  0 14  6)   ;  2
	 #(20 16 12 19 14  0  9)   ;  3
	 #(10  9  3  9  6  9  0))  ;  4
      (+ x 2))
     (+ y 2))))

(define vnd-ex-cap 25)

(define vnd-ex-avg-base-load 10)

(define vnd-ex-load-f
  (lambda (x)
    (vector-ref
     ;  -2 -1
     '#(13  6)
     (+ x 2))))

(define vnd-ex-test
  (lambda ()
    (vnd-solve vnd-ex-bs
	       vnd-ex-ds
	       vnd-ex-cost-f
	       vnd-ex-cap
	       vnd-ex-avg-base-load
	       vnd-ex-load-f)))

;------------------------------------------------------------------------------

; Implantation d'une heuristique pour resoudre le probleme du
; traveling salesman

; Recherche d'un CHEMIN a partir d'un certain point qui passe par
; toutes les home bases et qui est le moins couteux possible

; Heuristique:
; - Etape intermediaire: a partir d'un chemin reliant les home les
;   plus problematiques, choisir la meilleur insertion pour le home le
;   plus problematique
; - Etape initiale: fixer le premier chemin comme etant uniquement le
;   point initial
; - Maintenir un tableau indiquant la meilleure insertion pour chaque
;   home non-schedule

; Resolution symbolique: point initial: 0; homes: 1, ..., n; couts des
;   liens fournis par une fonction: node * node -> cost; solution:
;   (node ...) dont le premier est 0

; Representation concrete: tableau[0 .. n] de (couts ou #f), un cout
;   est le cout de l'insertion optimale, #f indique que le noeud est
;   deja schedule; tableau[0 .. n] de node*node, liste doublement
;   chainee du chemin ou lieu d'insertion optimal

(define tsp-solve
  (lambda (n cost)
    (set! tsp-n n)
    (set! tsp-cost cost)
    (tsp-init!)
    (let loop ((step 1))
      (if (<= step n)
	  (let* ((i (tsp-furthest))
		 (site (tsp-insert! i)))
	    (tsp-update! i site)
	    (loop (+ step 1)))
	  (tsp-extract-list)))))

; Variables d'information
(define tsp-n #f)
(define tsp-cost #f)
; Variables de travail
(define tsp-est #f)
(define tsp-link #f)

; (void) -> (void)
(define tsp-init!
  (lambda ()
    (set! tsp-est (make-vector (+ tsp-n 1) 200000000))
    (set! tsp-link (make-vector (+ tsp-n 1) #f))
    (vector-set! tsp-est 0 #f)                      ; 0 est deja dans le chemin
    (vector-set! tsp-link 0 (cons #f #f))                       ; aucun voisin!
    (let loop ((i 1))
      (if (<= i tsp-n)
	  (begin
	    (vector-set! tsp-est i (tsp-cost 0 i))
	    (vector-set! tsp-link i (cons 0 #f))        ; site: entre 0 et rien
	    (loop (+ i 1)))))))

; (void) -> node
(define tsp-furthest
  (lambda ()
    (let loop ((i 1) (max-dist -1) (max-i 0))
      (if (<= i tsp-n)
	  (let ((est (vector-ref tsp-est i)))
	    (if est
		(if (> est max-dist)
		    (loop (+ i 1) est i)
		    (loop (+ i 1) max-dist max-i))
		(loop (+ i 1) max-dist max-i)))
	  max-i))))

; node -> node * node
(define tsp-insert!
  (lambda (i)
    (let* ((site (vector-ref tsp-link i))                      ; site = (l . r)
	   (l (car site))
	   (r (cdr site))
	   (l-link (vector-ref tsp-link l))
	   (l-link (cons (car l-link) i)))
      (vector-set! tsp-link l l-link)                       ; fixe le lien de l
      (if r
	  (let* ((r-link (vector-ref tsp-link r))
		 (r-link (cons i (cdr r-link))))
	    (vector-set! tsp-link r r-link)))               ; fixe le lien de r
      (vector-set! tsp-est i #f)                           ; fixe l'estime de i
      site)))

; node , node * node -> (void)
(define tsp-update!
  (lambda (i site)
    (let* ((l (car site))
	   (r (cdr site)))
      (let loop ((j 1))
	(if (<= j tsp-n)
	    (let* ((est0 (vector-ref tsp-est j))   ; cout de l'ancien site de j
		   (est0 (if (and est0 (equal? (vector-ref tsp-link j) site))
			     200000000                ; site maintenant detruit
			     est0)))                     ; site encore existent
	      (if est0
		  (let* ((est1 (- (+ (tsp-cost l j)             ; cout de l-j-i
				     (tsp-cost j i))
				  (tsp-cost l i)))
			 (est2 (if r                            ; cout de i-j-r
				   (- (+ (tsp-cost i j)
					 (tsp-cost j r))
				      (tsp-cost i r))
				   (tsp-cost i j))))
		    (if (< (min est1 est2) est0)
			(if (< est1 est2)
			    (begin                              ; choisir l-j-i
			      (vector-set! tsp-est j est1)
			      (vector-set! tsp-link j (cons l i)))
			    (begin                              ; choisir i-j-r
			      (vector-set! tsp-est j est2)
			      (vector-set! tsp-link j (cons i r)))))))
	      (loop (+ j 1))))))))

; (void) -> [list of node]
(define tsp-extract-list
  (lambda ()
    (let loop ((i 0))
      (if i
	  (cons i (loop (cdr (vector-ref tsp-link i))))
	  '()))))

; Donnees pour exemple
(define tsp-example-data
  '((0  15 25 50 10 20)
    (15  0 30 55 25 15)
    (25 30  0 30 30 10)
    (50 55 30  0 45 40)
    (10 25 30 45  0 20)
    (20 15 10 40 20  0)))

; Fonction "cost" pour exemple
(define tsp-example-cost-1
  (lambda (i j)
    (list-ref (list-ref tsp-example-data i) j)))

(define discrete-sqrt
  (lambda (n)
    (let loop1 ((r 1))
      (if (<= (* r r) n)
	  (loop1 (* 2 r))
	  (let loop2 ((low (if (even? r) (quotient r 2) 0)) (high r))
	    (if (<= (- high low) 1)
		low
		(let ((med (quotient (+ low high) 2)))
		  (if (<= (* med med) n)
		      (loop2 med high)
		      (loop2 low med)))))))))

; Fonction "cost" pour exemple.  Grille 10x10; (10,10) ... (100,100);
; partant de (0,10); cout optimal du chemin = 1000
(define tsp-example-cost-2
  (lambda (m n)
    (let* ((pos
	    (lambda (n)
	      (if (= n 0)
		  (list 0 10)
		  (list (* (+ (modulo (- n 1) 10) 1) 10)                 ; en x
			(* (+ (quotient (- n 1) 10) 1) 10)))))           ; en y
	   (mpos (pos m))
	   (npos (pos n)))
      (discrete-sqrt (+ (expt (- (car mpos) (car npos)) 2)
			(expt (- (cadr mpos) (cadr npos)) 2))))))

(define tsp-path-cost
  (lambda (path cost)
    (let loop ((i (car path)) (path (cdr path)) (len 0))
      (if (null? path)
	  len
	  (let ((j (car path)))
	    (loop j (cdr path) (+ len (cost i j))))))))

;------------------------------------------------------------------------------

(define (expert-avoid-drowning port)

  (define expert-name 'avoid-drowning)

  (let ((pos (robot-pos me)))
    (if (next-to-water-and-pushable? pos rid)
        (let ((dirs (possible-directions pos)))
          (apply
           append
           (map (lambda (dir)
                  (desire-urgent-Move dir 100))
                dirs)))
        (desire-nothing))))

;------------------------------------------------------------------------------

(define (expert-drown-enemy port)

  (define expert-name 'drown-enemy)

  (let ((pos (robot-pos me)))
    (let ((dirs (possible-directions pos)))
      (apply
       append
       (map (lambda (dir)
              (let ((neighbor (displace pos dir)))
                (let ((r (pos->robot neighbor)))
                  (if (and r
                           (next-to-water-and-pushable? neighbor (robot-id r)))
                      (desire-urgent-Move dir 95)
                      '()))))
            dirs)))))

;------------------------------------------------------------------------------

(define asoc-count 0)
(define asoc? #t)

(define-learnable-parameter asociability-min-dist 0 15)

(define (expert-asociability port)

  (define expert-name 'asociability)

  (let* ((pos (robot-pos me))
         (c (control pos control-map-excluding-me)))

    ; c is a list like (<distance> <robot-id1> <robot-id2>...)
    ; - if we are alone in the world the distance is infinite (actually
    ;   9999999) and there is nothing after the distance
    ; - there is more than one robot-id if they are equidistant

    (let ((distance (car c)))

      (if (and asoc?
               (< distance (asociability-min-dist)))

          (let ((dirs (possible-directions pos)))
            (if (null? dirs)

                (begin
                  (set! asoc-count 0)
                  (desire-nothing))

                (let* ((x
                        (choose-highest
                         (map (lambda (dir)
                                (let ((dist
                                       (car (control
                                             (displace pos dir)
                                             control-map-excluding-me))))
                                  (cons dist dir)))
                              dirs)))
                       (highest-dist
                        (car x))
                       (highest-dir
                        (cdr x)))
                  (if (> highest-dist distance)

                      (begin
                        (set! asoc-count (+ asoc-count (rand 2)))
                        (if (> asoc-count 3)
                            (set! asoc? #f))
                        (desire-Move highest-dir 90))

                      (begin
                        (set! asoc-count 0)
                        (desire-nothing))))))

          (begin
            (if (not asoc?)
                (begin
                  (set! asoc-count (- asoc-count (rand 2)))
                  (if (< asoc-count 1)
                      (set! asoc? #t))))
            (desire-nothing))))))

;------------------------------------------------------------------------------

(define (expert-delivery1 port)

  (define expert-name 'delivery1)

  (let* ((pos (robot-pos me))
         (my-packages (robot-packages me)))
    (if (null? my-packages) ; we aren't carrying packages

        (desire-nothing)

        (let ((lst (keep (lambda (id)
                           (let ((p (id->package id)))
                             (= (package-pos p) pos)))
                         my-packages)))
          (if (null? lst) ; no packages are destined here

              (desire-nothing)

              (desire-Drop lst 85))))))

;------------------------------------------------------------------------------

'(define-learnable-parameter delivery2-distance-horizon 0 15)
'
(define (expert-delivery2 port)

  (define expert-name 'delivery2)

  (let* ((pos (robot-pos me))
         (my-packages (robot-packages me)))
    (if (null? my-packages) ; we aren't carrying packages

        (desire-nothing)

        (let ((closest-destination
               (closest (map (lambda (id)
                               (let ((p (id->package id)))
                                 (package-pos p)))
                             my-packages))))

          (if (<= (distance-to closest-destination
                               distance-map-from-me)
                  (delivery2-distance-horizon))

              (desire-Move (direction-to closest-destination) 80)

              (desire-nothing))))))

;------------------------------------------------------------------------------

(define (expert-delivery3 port)

  (define expert-name 'delivery3)

  (let* ((pos (robot-pos me))
         (my-packages (robot-packages me)))
    (if (null? my-packages) ; we aren't carrying packages

        (desire-nothing)

        (let ((closest-destination
               (closest (map (lambda (id)
                               (let ((p (id->package id)))
                                 (package-pos p)))
                             my-packages))))

          (if (< (distance-to closest-destination
                              distance-map-from-me)
                 infinite)

              (desire-Move (direction-to closest-destination) 75)

              (desire-nothing))))))

;------------------------------------------------------------------------------

(define (expert-greediness port)

  (define expert-name 'greediness)

  (let* ((pos (robot-pos me))
         (lst (packages-known-to-be-pickable pos)))
    (if (null? lst) ; no packages can be picked!

        (desire-nothing)

        (desire-Pick lst 70))))

;------------------------------------------------------------------------------

(define (expert-transport port)
  (define expert-name 'transport)
  (set! RO-path (cleanup-RO-path RO-path))
  (if (null? RO-path)
      (begin
	(set! RO-path (cleanup-RO-path (visit-and-deliver)))
	(if (null? RO-path)
	    (let ((dirs (possible-directions (robot-pos me))))
	      (if (null? dirs)
		  (desire-Pick '() 65)
		  (desire-Move (car dirs) 65)))
	    (desire-Move (direction-to (car RO-path)) 65)))
      (desire-Move (direction-to (car RO-path)) 65)))

;------------------------------------------------------------------------------

'(define-learnable-parameter curiosity1-distance-horizon 0 15)
'
(define (expert-curiosity1 port)

  (define expert-name 'curiosity1)

  (let ((interesting-positions
         (keep (lambda (base-pos)
                 (let ((p (pos->packages base-pos)))
                   (and (or (uncertain? p)
                            (not (null?
                                  (packages-known-to-be-pickable base-pos))))
                        (<= (distance-to base-pos distance-map-from-me)
                            (curiosity1-distance-horizon)))))
               bases)))
    (if (null? interesting-positions)
        (desire-nothing)
        (let ((dir
               (direction-to
                (closest interesting-positions))))
          (desire-Move dir 60)))))

;------------------------------------------------------------------------------

(define (expert-curiosity2 port)

  (define expert-name 'curiosity2)

  (let ((interesting-positions
         (keep (lambda (base-pos)
                 (let ((p (pos->packages base-pos)))
                   (and (or (uncertain? p)
                            (not (null?
                                  (packages-known-to-be-pickable base-pos))))
                        (< (distance-to base-pos distance-map-from-me)
                            infinite))))
               bases)))
    (if (null? interesting-positions)
        (desire-nothing)
        (let ((dir
               (direction-to
                (closest interesting-positions))))
          (desire-Move dir 55)))))

;------------------------------------------------------------------------------

'(define-learnable-parameter curiosity3-cell-horizon 0 5)
'(define-learnable-parameter curiosity3-distance-horizon 0 15)
'
(define (expert-curiosity3 port)

  (define expert-name 'curiosity3)

  (let ((interesting-positions
         (scan (robot-pos me)
               (curiosity3-cell-horizon)
               '()
               (lambda (pos lst)
                 (if (and (<= (distance-to pos distance-map-from-me)
                              (curiosity3-distance-horizon))
                          (or (uncertain? (pos->packages pos))
                              (not (null?
                                    (packages-known-to-be-pickable pos)))))
                     (cons pos lst)
                     lst)))))
    (if (null? interesting-positions)
        (desire-nothing)
        (let ((dir
               (direction-to
                (closest interesting-positions))))
          (desire-Move dir 50)))))

;------------------------------------------------------------------------------

(define (expert-dropper port)
  (define expert-name 'dropper)
  (let ((lst (packages-brought-to-dest)))
    (if (null? lst) ; no packages at dest
	(desire-nothing)
	(desire-Drop lst 85))))

;------------------------------------------------------------------------------

(define (think-and-send-command port)
  (let* ((actions
          (append (expert-avoid-drowning port)
                  (expert-drown-enemy port)
                  (expert-asociability port)
                  (expert-dropper port)
;                  (expert-delivery1 port)
;                  (expert-delivery2 port)
;                  (expert-delivery3 port)
                  (expert-greediness port)
;                  (expert-curiosity1 port)
;                  (expert-curiosity2 port)
;                  (expert-curiosity3 port)
                  (expert-transport port)
                  (list (make-Pick 0 1 '())))) ; no-op
         (combined
          (combine-actions actions)))

    (if trace?
        (begin
          (pp '=================================================)
          (pp actions)
          (pp combined)
	  (RO-trace)))

    (let ((act (car combined)))

      ; Hooks: sert a faire des choses
      (cond ((Move? act)
	     'none)
	    ((Pick? act)
	     (set! RO-path #f))
	    ((Drop? act)
	     (set! RO-path #f))
	    (else
	     (error "unknown action")))

      (write-action act port)
      (pay (action-bid act) port))))

;------------------------------------------------------------------------------

;(trace set-robot-at! remove-robot-at! pos->robot)

(apply main (cdr (command-line)))

(force-output (current-output-port)) ; **** IMPORTANT ****
